﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Security.Cryptography;
using System.IO;
using CaptainQuery.Utils;

namespace CaptainQuery.CryptoHashProvider
{
    /// <summary>
    /// A wrapper around a ICryptoTransForm implementation
    /// which adds hash functionality 
    /// </summary>
    public class HashedCryptoTransform : ICryptoTransform
    {
        bool _encrypt;
        MemoryStream _ms = null;
        ICryptoTransform _innerCryptoTransForm = null;
        byte[] _hashSecret = null;

        void RecycleMemorystream(byte[] bytes = null)
        {
            // Clear and dispose the memory stream
            var l = _ms.Length;

            for (int i = 0; i < l; i++) _ms.WriteByte(0);
            _ms.Clear();
            
            //_ms.Clear();
            _ms.Dispose();
            // Create a new stream
            if (bytes != null)
            {
                _ms = new MemoryStream();
                _ms.Write(bytes,0,bytes.Length);
            }
            else
                _ms = new MemoryStream();
            

        }

        public HashedCryptoTransform(bool encrypt, ICryptoTransform innerCryptoTransform)
        {
            _ms = new MemoryStream();
            this._hashSecret = Config.CryptoSHASecret;
            this._innerCryptoTransForm = innerCryptoTransform;
            this._encrypt = encrypt;
        }

        public bool CanReuseTransform
        {
            get { return _innerCryptoTransForm.CanReuseTransform; }
        }

        public bool CanTransformMultipleBlocks
        {
            get { return _innerCryptoTransForm.CanTransformMultipleBlocks; }
        }

        public int InputBlockSize
        {
            get { return _innerCryptoTransForm.InputBlockSize; }
        }

        public int OutputBlockSize
        {
            get { return _innerCryptoTransForm.OutputBlockSize; }
        }

        public int TransformBlock(byte[] inputBuffer, int inputOffset, int inputCount, byte[] outputBuffer, int outputOffset)
        {
            // buffer the request        
            _ms.Write(inputBuffer, inputOffset, inputCount);
            // do not advance the output
            return outputOffset;
        }


        public byte[] TransformFinalBlock(byte[] inputBuffer, int inputOffset, int inputCount)
        {
            try
            {
                // buffer the request
                _ms.Write(inputBuffer, inputOffset, inputCount);
                if (_encrypt)
                    return EncryptFinalBlock();
                else
                    return DecryptFinalBlock();
            }
            finally
            {
                RecycleMemorystream();
            }
        }

        private byte[] DecryptFinalBlock()
        {
            using (var sha = new SHA256Managed())
            {
                // get the Hash in the encrypted stream
                var hashBytes = (sha.HashSize / 8);
                byte[] storedHash = _ms.GetBuffer().SubArray((int)_ms.Length - hashBytes, hashBytes);
                // backup the stream and add the hashsecret
                _ms.Seek(-(hashBytes), SeekOrigin.End);
                _ms.SetLength(_ms.Position);
                _ms.Write(_hashSecret, 0, _hashSecret.Length);
                // compute the hash
                byte[] computedHash = sha.ComputeHash(_ms.GetBuffer(), 0, (int)_ms.Length);
                // compare the hash

                if (!computedHash.CompareTo(storedHash))
                    throw new InvalidHashException();                       
                       
                // decrypt the data
                byte[] ret = _innerCryptoTransForm.TransformFinalBlock(_ms.GetBuffer(), 0, (int)_ms.Length - _hashSecret.Length);
                return ret;
            }

        }

        private byte[] EncryptFinalBlock()
        {
            using (var sha = new SHA256Managed())
            {
                // encrypt the data and recycle the memory stram with this data
                byte[] newdata = _innerCryptoTransForm.TransformFinalBlock(_ms.GetBuffer(), 0, (int)_ms.Length);
                RecycleMemorystream(newdata);

                // add the hashsecret and compute the hash
                var pos = _ms.Length;
                _ms.Write(_hashSecret, 0, _hashSecret.Length);
                byte[] hash = sha.ComputeHash(_ms.GetBuffer(), 0, (int)_ms.Length);

                // backup the stream, remove the hash secret and add the hash
                _ms.Seek(pos, SeekOrigin.Begin);
                _ms.SetLength(pos);
                _ms.Write(hash, 0, hash.Length);
                // return the data
                return _ms.GetBuffer().SubArray(0, (int)_ms.Length);
            }
        }

        public void Dispose()
        {
            if (_innerCryptoTransForm != null) { _innerCryptoTransForm.Dispose(); _innerCryptoTransForm = null; }
            if (_ms != null) { _ms.Dispose(); _ms = null; }
           

        }        
    }
}
